<?php

if (class_exists('PEAR_Sniffs_Classes_ClassDeclarationSniff', true) === false) {
    $error = 'Class PEAR_Sniffs_Classes_ClassDeclarationSniff not found';
    throw new PHP_CodeSniffer_Exception($error);
}

class Maarch_Sniffs_Classes_ClassDeclarationSniff
    extends PEAR_Sniffs_Classes_ClassDeclarationSniff
{


    /**
     * Returns an array of tokens this test wants to listen for.
     *
     * @return array
     */
    public function register()
    {
        return array(
            T_CLASS,
            T_INTERFACE,
        );

    }//end register()


    /**
     * Processes this test, when one of its tokens is encountered.
     *
     * @param PHP_CodeSniffer_File $phpcsFile The file being scanned.
     * @param int                  $stackPtr  The position of the current token
     *                                         in the stack passed in $tokens.
     *
     * @return void
     */
    public function process(PHP_CodeSniffer_File $phpcsFile, $stackPtr)
    {
        // We want all the errors from the PEAR standard, plus some of our own.
        parent::process($phpcsFile, $stackPtr);

        $tokens = $phpcsFile->getTokens();

        /*
            Check alignment of the keyword and braces.
        */

        if ($tokens[($stackPtr - 1)]['code'] === T_WHITESPACE) {
            $prevContent = $tokens[($stackPtr - 1)]['content'];
            if ($prevContent !== $phpcsFile->eolChar) {
                $blankSpace = substr($prevContent, 
                                     strpos($prevContent, $phpcsFile->eolChar));
                $spaces = strlen($blankSpace);

                if (
                    in_array($tokens[($stackPtr - 2)]['code'], 
                             array(T_ABSTRACT, T_FINAL)) === true
                ) {
                    if ($spaces !== 1) {
                        $type = strtolower($tokens[$stackPtr]['content']);
                        $prevContent = strtolower(
                            $tokens[($stackPtr - 2)]['content']
                        );
                        $error = 'Expected 1 space between %s and %s keywords; '
                               . '%s found';
                        $data = array(
                            $prevContent,
                            $type,
                            $spaces,
                        );
                        $phpcsFile->addError($error, $stackPtr - 1, 
                        					 'SpacesBeforeKeyword', $data);
                    }
                }
            }
        }//end if

        if (isset($tokens[$stackPtr]['scope_opener']) === false) {
            $error = 'Possible parse error: %s missing opening or '
                   . 'closing brace';
            $data  = array($tokens[$stackPtr]['content']);
            $phpcsFile->addWarning($error, $stackPtr, 'MissingBrace', $data);
            return;
        }

        $closeBrace = $tokens[$stackPtr]['scope_closer'];
        if ($tokens[($closeBrace - 1)]['code'] === T_WHITESPACE) {
            $prevContent = $tokens[($closeBrace - 1)]['content'];
            if ($prevContent !== $phpcsFile->eolChar) {
                $blankSpace = substr($prevContent, 
                                     strpos($prevContent, $phpcsFile->eolChar));
                $spaces = strlen($blankSpace);
                $spacesLine = $tokens[($closeBrace - 1)]['line'];
                if ($spaces !== 0
                    && ($spacesLine === $tokens[$closeBrace]['line'])
                ) {
                        $error = 'Expected 0 spaces before closing brace; '
                               . '%s found';
                        $data  = array($spaces);
                        $phpcsFile->addError($error, $closeBrace - 1, 
                                             'SpaceBeforeCloseBrace', $data);
                }
            }
        }

        /*
            Check that each of the parent classes or interfaces specified
            are spaced correctly.
        */

        // We need to map out each of the possible tokens in the declaration.
        $keyword      = $stackPtr;
        $openingBrace = $tokens[$stackPtr]['scope_opener'];
        $className    = $phpcsFile->findNext(T_STRING, $stackPtr);
        $keywordLine = $tokens[$keyword]['line'];
        $openingBraceLine = $tokens[$openingBrace]['line'];
        $isMultiline = ($keywordLine !== ($openingBraceLine - 1));

        /*
            Now check the spacing of each token.
        */

        $name = strtolower($tokens[$keyword]['content']);

        // Spacing of the keyword.
        $gap = $tokens[($stackPtr + 1)]['content'];
        if (strlen($gap) !== 1) {
            $found = strlen($gap);
            $error = 'Expected 1 space between %s keyword and %s name; '
                   . '%s found';
            $data  = array($name, $name, $found);
            $phpcsFile->addError($error, $stackPtr + 1, 
            					 'SpaceAfterKeyword', $data);
        }


        // Now check each of the parents.
        $parents    = array();
        $nextParent = ($className + 1);
        while (
            ($nextParent = $phpcsFile->findNext(
                array(T_IMPLEMENTS, T_EXTENDS), 
                ($nextParent + 1), 
                ($openingBrace)
            )) !== false
        ) {
            $parents[] = $nextParent;
        }
        
        $parentCount = count($parents);
        $previousLine = $keywordLine;

        for ($i = 0; $i < $parentCount; $i++) {
            $curToken = $parents[$i];
            $tokenName = $tokens[$curToken]['content'];

            $this->_checkSpacing($phpcsFile, $curToken);
            if (!$isMultiline) {
                $this->_checkSingleLineSpacing($phpcsFile, $curToken);
            }
             
            
            if ($isMultiline) {
                if ($tokens[$curToken]['line'] === $previousLine) {
                    // if keyword is on the same line, raise warning
                    $warning = '%s should be on next line';
                    $data = array($tokenName);
                    $phpcsFile->addWarning($warning, $curToken,
                                           'KeywordOnNextLine', $data);
                } else {
                    // if it is on the next line
                    $previousLine = $tokens[$curToken]['line'];
                    
                    //check indent
                    $firstOnLine = $phpcsFile->findFirstOnLine(
                        array(T_WHITESPACE), $curToken
                    );
                    if ($firstOnLine === false) {
                        $error = 'Keyword %s must be indented';
                        $data = array($tokenName);
                        $phpcsFile->addError($error, $curToken,
                                             'BadIndent', $data);
                    } else {
                        $indentLength = strlen(
                            $tokens[$firstOnLine]['content']
                        );
                        if ($indentLength !== 4) {
                            $error = 'Keyword %s must be indented 4 spaces;'
                                   . '%s found';
                            $data = array($tokenName, $indentLength);
                            $phpcsFile->addError($error, $curToken - 1,
                                                 'BadIndent', $data);
                        }
                    }
                }
                
                // Check indent of interfaces
                if ($tokens[$curToken]['code'] === T_IMPLEMENTS) {
                    $this->_checkInterfaceIndent(
                        $phpcsFile, $curToken, $openingBrace
                    );
                    $this->_checkInterfaceCommaSpacing(
                        $phpcsFile, $curToken, $openingBrace
                    );
                }
            }            
            
            /*if ($tokens[$parents[$i]]['code'] === T_IMPLEMENTS) {
                continue;
            }

            if ($tokens[($parents[$i] - 1)]['code'] !== T_WHITESPACE) {
                $name  = $tokens[$parents[$i]]['content'];
                $error = 'Expected 1 space before "%s"; 0 found';
                $data  = array($name);
                $phpcsFile->addError($error, ($nextComma + 1), 
                					 'NoSpaceBeforeName', $data);
            } else {
                $spaceBefore = strlen($tokens[($parents[$i] - 1)]['content']);
                if ($spaceBefore !== 1) {
                    $name  = $tokens[$parents[$i]]['content'];
                    $error = 'Expected 1 space before "%s"; %s found';
                    $data  = array($name, $spaceBefore);
                    $phpcsFile->addError($error, $stackPtr, 
                    					 'SpaceBeforeName', $data);
                }
            }

            if ($tokens[($parents[$i] + 1)]['code'] !== T_COMMA) {
                if ($i !== ($parentCount - 1)) {
                    // This is not the last parent, and the comma
                    // is not where we expect it to be.
                    if ($tokens[($parents[$i] + 2)]['code'] !== T_IMPLEMENTS) {
                        $found = strlen($tokens[($parents[$i] + 1)]['content']);
                        $name  = $tokens[$parents[$i]]['content'];
                        $error = 'Expected 0 spaces between "%s" and comma; '
                               . '$%s found';
                        $data  = array($name, $found);
                        $phpcsFile->addError($error, $stackPtr, 'SpaceBeforeComma', $data);
                    }
                }

                $nextComma = $phpcsFile->findNext(T_COMMA, $parents[$i]);
            } else {
                $nextComma = ($parents[$i] + 1);
            }*/
        }//end for*/

    }//end process()
    
    private function _checkSpacing($phpcsFile, $curToken) 
    {
        $tokens = $phpcsFile->getTokens();
        
        $tokenName = $tokens[$curToken]['content'];
        $spaceBeforeLength = strlen($tokens[$curToken - 1]['content']);
        $spaceAfterLength = strlen($tokens[$curToken + 1]['content']);
        
        if ($spaceAfterLength !== 1) {
            $error = 'Expected one space between %s and %s; '
                   . 'found %s';
            $data = array(
                $tokenName, 
                $tokens[$curToken + 2]['content'], 
                $spaceAfterLength,
            );
            $phpcsFile->addError($error, $curToken + 1,
                                 'SpaceAfterKeyword', $data);
        }
    }
    
    private function _checkSingleLineSpacing($phpcsFile, $curToken) 
    {
        $tokens = $phpcsFile->getTokens();
        $tokenName = $tokens[$curToken]['content'];
        $spaceBeforeLength = strlen($tokens[$curToken - 1]['content']);
        $spaceAfterLength = strlen($tokens[$curToken + 1]['content']);
        
        if ($spaceBeforeLength !== 1) {
            $error = 'Expected one space between %s and %s; '
                   . 'found %s';
            $data = array(
                $tokens[$curToken - 2]['content'], 
                $tokenName, 
                $spaceBeforeLength,
            );
            $phpcsFile->addError($error, $curToken - 1,
                                 'SpaceBeforeKeyword', $data);
        }
    }
    
    private function _checkInterfaceCommaSpacing(
        $phpcsFile, $curToken, $openingBrace
    ) {
        $tokens = $phpcsFile->getTokens();
        $nextComma = $curToken;
        while (
            ($nextComma = $phpcsFile->findNext(
                Array(T_COMMA), 
                $nextComma + 1, 
                $openingBrace
            )) !== false
        ) {
            $nextToken = $nextComma + 1;
            $prevToken = $nextComma - 1;
            if ($tokens[$prevToken]['code'] === T_WHITESPACE) {
                $error = 'There must be no space before comma';
                $phpcsFile->addError($error, $prevToken,
                                     'NoSpaceBeforeComma');
            }
            
            $nextNonEmptyToken = $phpcsFile->findNext(
                array(T_WHITESPACE, T_COMMENT),
                $nextComma + 1, $openingBrace, true
            );
            $nextNonEmptyTokenLine = $tokens[$nextNonEmptyToken]['line'];
            $nextCommaLine = $tokens[$nextComma]['line'];
            $isEndOfLine = ($nextNonEmptyTokenLine !== $nextCommaLine);
            
            if (!$isEndOfLine){
                if ($tokens[$nextToken]['code'] !== T_WHITESPACE) {
                    $error = 'There must be one space after comma';
                    $phpcsFile->addError($error, $nextComma,
                                         'SpaceAfterComma');
                } else {
                    // only if not last on line
                    $spaceLength = strlen($tokens[$nextToken]['content']);
                    if ($spaceLength !== 1) {
                        $error = 'There must be a single space after comma';
                        $phpcsFile->addError($error, $nextComma,
                                             'OneSpaceAfterComma');
                    }
                }
            }
        }
    }
    
    private function _checkInterfaceIndent(
        $phpcsFile, $curToken, $openingBrace
    ) {
        $tokens = $phpcsFile->getTokens();
        $whileLine = $tokens[$curToken]['line'];
        // $expectedIndent = 4 + len("implements") + 1
        $expectedIndent = 15;
        $nextIface = $curToken;
        while (
            ($nextIface = $phpcsFile->findNext(
                Array(T_STRING), 
                $nextIface + 1, 
                $openingBrace
            )) !== false
        ) {
            //we only care about first token on the line
            if ($tokens[$nextIface]['line'] === $whileLine) {
                continue;
            }
            
            $whileLine = $tokens[$nextIface]['line'];
            $ifaceIndent = $phpcsFile->findFirstOnLine(
                Array(T_WHITESPACE), $nextIface
            );
            if ($ifaceIndent === false) {
                $error = 'Interface name must be indented'
                       . '%s spaces';
                $data = array($expectedIndent);
                $phpcsFile->addError($error, $nextIface,
                                     'BadIndent', $data);
            } else {
                $indentLength = strlen($tokens[$ifaceIndent]['content']);
                if ($indentLength !== $expectedIndent) {
                    $error = 'Interface name must be indented'
                       . '%s spaces; found %s';
                    $data = array($expectedIndent, $indentLength);
                    $phpcsFile->addError($error, $nextIface,
                                         'BadIndent', $data);
                }
            }
        }
    }


}//end class